"""
Tests for 'sourmash sig collect'
"""

import pytest
import shutil
import os.path
import gzip

import sourmash
from sourmash.manifest import BaseCollectionManifest

import sourmash_tst_utils as utils
from sourmash_tst_utils import SourmashCommandFailed


def test_sig_collect_0_nothing(runtmp, manifest_db_format, abspath_relpath_v4):
    # run with just output
    ext = "sqlmf" if manifest_db_format == "sql" else "csv"
    if manifest_db_format != "sql":
        return

    runtmp.sourmash(
        "sig",
        "collect",
        "-o",
        f"mf.{ext}",
        "-F",
        manifest_db_format,
        abspath_relpath_v4,
    )

    manifest_fn = runtmp.output(f"mf.{ext}")
    manifest = BaseCollectionManifest.load_from_filename(manifest_fn)

    assert len(manifest) == 0


def test_sig_collect_0_fail_abspath_relpath(runtmp, manifest_db_format):
    # check that it complains if both --abspath and --relpath are specified
    ext = "sqlmf" if manifest_db_format == "sql" else "csv"
    if manifest_db_format != "sql":
        return

    with pytest.raises(
        SourmashCommandFailed,
        match="Cannot specify both --abspath and --relpath; pick one!",
    ):
        runtmp.sourmash(
            "sig",
            "collect",
            "-o",
            f"mf.{ext}",
            "-F",
            manifest_db_format,
            "--abspath",
            "--relpath",
        )


def test_sig_collect_1_zipfile(runtmp, manifest_db_format, abspath_relpath_v4):
    # collect a manifest from a .zip file
    protzip = utils.get_test_data("prot/protein.zip")

    ext = "sqlmf" if manifest_db_format == "sql" else "csv"

    runtmp.sourmash(
        "sig",
        "collect",
        protzip,
        "-o",
        f"mf.{ext}",
        "-F",
        manifest_db_format,
        abspath_relpath_v4,
    )

    manifest_fn = runtmp.output(f"mf.{ext}")
    manifest = BaseCollectionManifest.load_from_filename(manifest_fn)

    assert len(manifest) == 2
    md5_list = [row["md5"] for row in manifest.rows]
    assert "16869d2c8a1d29d1c8e56f5c561e585e" in md5_list
    assert "120d311cc785cc9d0df9dc0646b2b857" in md5_list


def test_sig_collect_1_zipfile_csv_gz(runtmp, abspath_relpath_v4):
    # collect a manifest from a .zip file, save to csv.gz
    protzip = utils.get_test_data("prot/protein.zip")

    runtmp.sourmash(
        "sig", "collect", protzip, "-o", "mf.csv.gz", "-F", "csv", abspath_relpath_v4
    )

    manifest_fn = runtmp.output("mf.csv.gz")

    # gzip, yes?
    print("XXX", manifest_fn)
    with gzip.open(manifest_fn, "rt", newline="") as fp:
        fp.read()

    manifest = BaseCollectionManifest.load_from_filename(manifest_fn)

    assert len(manifest) == 2
    md5_list = [row["md5"] for row in manifest.rows]
    assert "16869d2c8a1d29d1c8e56f5c561e585e" in md5_list
    assert "120d311cc785cc9d0df9dc0646b2b857" in md5_list


def test_sig_collect_1_zipfile_csv_gz_roundtrip(runtmp, abspath_relpath_v4):
    # collect a manifest from a .zip file, save to csv.gz; then load again
    protzip = utils.get_test_data("prot/protein.zip")

    runtmp.sourmash(
        "sig", "collect", protzip, "-o", "mf.csv.gz", "-F", "csv", abspath_relpath_v4
    )

    manifest_fn = runtmp.output("mf.csv.gz")

    # gzip, yes?
    print("XXX", manifest_fn)
    with gzip.open(manifest_fn, "rt", newline="") as fp:
        fp.read()

    manifest = BaseCollectionManifest.load_from_filename(manifest_fn)

    assert len(manifest) == 2
    md5_list = [row["md5"] for row in manifest.rows]
    assert "16869d2c8a1d29d1c8e56f5c561e585e" in md5_list
    assert "120d311cc785cc9d0df9dc0646b2b857" in md5_list

    # can we read a csv.gz?
    runtmp.sourmash("sig", "collect", "mf.csv.gz", "-o", "mf2.csv", "-F", "csv")

    manifest_fn2 = runtmp.output("mf2.csv")
    manifest2 = BaseCollectionManifest.load_from_filename(manifest_fn2)

    assert len(manifest2) == 2
    md5_list = [row["md5"] for row in manifest2.rows]
    assert "16869d2c8a1d29d1c8e56f5c561e585e" in md5_list
    assert "120d311cc785cc9d0df9dc0646b2b857" in md5_list


def test_sig_collect_2_exists_fail(runtmp, manifest_db_format):
    # collect a manifest from two .zip files
    protzip = utils.get_test_data("prot/protein.zip")
    allzip = utils.get_test_data("prot/protein.zip")

    ext = "sqlmf" if manifest_db_format == "sql" else "csv"

    runtmp.sourmash(
        "sig", "collect", protzip, "-o", f"mf.{ext}", "-F", manifest_db_format
    )

    manifest_fn = runtmp.output(f"mf.{ext}")
    manifest = BaseCollectionManifest.load_from_filename(manifest_fn)

    assert len(manifest) == 2
    md5_list = [row["md5"] for row in manifest.rows]
    assert "16869d2c8a1d29d1c8e56f5c561e585e" in md5_list
    assert "120d311cc785cc9d0df9dc0646b2b857" in md5_list

    # now run with same filename - should fail
    with pytest.raises(SourmashCommandFailed):
        runtmp.sourmash(
            "sig", "collect", allzip, "-o", manifest_fn, "-F", manifest_db_format
        )


def test_sig_collect_2_exists_merge(runtmp, manifest_db_format, abspath_relpath_v4):
    # collect a manifest from two .zip files
    protzip = utils.get_test_data("prot/protein.zip")
    allzip = utils.get_test_data("prot/all.zip")

    ext = "sqlmf" if manifest_db_format == "sql" else "csv"

    runtmp.sourmash(
        "sig",
        "collect",
        protzip,
        "-o",
        f"mf.{ext}",
        "-F",
        manifest_db_format,
        abspath_relpath_v4,
    )

    manifest_fn = runtmp.output(f"mf.{ext}")
    manifest = BaseCollectionManifest.load_from_filename(manifest_fn)

    assert len(manifest) == 2
    md5_list = [row["md5"] for row in manifest.rows]
    assert "16869d2c8a1d29d1c8e56f5c561e585e" in md5_list
    assert "120d311cc785cc9d0df9dc0646b2b857" in md5_list

    # now run with same filename - should merge
    runtmp.sourmash(
        "sig", "collect", allzip, "-o", manifest_fn, "-F", manifest_db_format, "--merge"
    )

    manifest = BaseCollectionManifest.load_from_filename(manifest_fn)
    assert len(manifest) == 10


def test_sig_collect_2_exists_sql_merge_csv(runtmp, manifest_db_format):
    # try to merge csv into sql
    protzip = utils.get_test_data("prot/protein.zip")
    allzip = utils.get_test_data("prot/all.zip")

    ext = "sqlmf"

    # save as sql...
    runtmp.sourmash("sig", "collect", protzip, "-o", f"mf.{ext}", "-F", "sql")

    manifest_fn = runtmp.output(f"mf.{ext}")
    manifest = BaseCollectionManifest.load_from_filename(manifest_fn)

    assert len(manifest) == 2
    md5_list = [row["md5"] for row in manifest.rows]
    assert "16869d2c8a1d29d1c8e56f5c561e585e" in md5_list
    assert "120d311cc785cc9d0df9dc0646b2b857" in md5_list

    with pytest.raises(SourmashCommandFailed):
        runtmp.sourmash(
            "sig", "collect", allzip, "-o", manifest_fn, "-F", "csv", "--merge"
        )

    assert "ERROR loading" in runtmp.last_result.err


def test_sig_collect_2_exists_csv_merge_sql(runtmp):
    # try to merge sql into csv
    protzip = utils.get_test_data("prot/protein.zip")
    allzip = utils.get_test_data("prot/all.zip")

    ext = "csv"

    # save as csv...
    runtmp.sourmash("sig", "collect", protzip, "-o", f"mf.{ext}", "-F", "csv")

    manifest_fn = runtmp.output(f"mf.{ext}")
    manifest = BaseCollectionManifest.load_from_filename(manifest_fn)

    assert len(manifest) == 2
    md5_list = [row["md5"] for row in manifest.rows]
    assert "16869d2c8a1d29d1c8e56f5c561e585e" in md5_list
    assert "120d311cc785cc9d0df9dc0646b2b857" in md5_list

    with pytest.raises(SourmashCommandFailed):
        runtmp.sourmash(
            "sig", "collect", allzip, "-o", manifest_fn, "-F", "sql", "--merge"
        )

    assert "ERROR loading" in runtmp.last_result.err


def test_sig_collect_2_no_exists_merge(runtmp, manifest_db_format, abspath_relpath_v4):
    # test 'merge' when args.output doesn't already exist => warning
    utils.get_test_data("prot/protein.zip")
    allzip = utils.get_test_data("prot/all.zip")

    ext = "sqlmf" if manifest_db_format == "sql" else "csv"
    manifest_fn = runtmp.output(f"mf.{ext}")

    # run with --merge but no previous:
    runtmp.sourmash(
        "sig",
        "collect",
        allzip,
        "-o",
        manifest_fn,
        "-F",
        manifest_db_format,
        "--merge",
        abspath_relpath_v4,
    )

    manifest = BaseCollectionManifest.load_from_filename(manifest_fn)
    assert len(manifest) == 8

    err = runtmp.last_result.err
    print(err)
    assert "WARNING: --merge-previous specified, but output file" in err


def test_sig_collect_3_multiple(runtmp, manifest_db_format):
    # collect a manifest from two .zip files
    protzip = utils.get_test_data("prot/protein.zip")
    hpzip = utils.get_test_data("prot/hp.zip")
    dayzip = utils.get_test_data("prot/dayhoff.zip")

    ext = "sqlmf" if manifest_db_format == "sql" else "csv"

    runtmp.sourmash(
        "sig",
        "collect",
        protzip,
        hpzip,
        dayzip,
        "-o",
        f"mf.{ext}",
        "-F",
        manifest_db_format,
    )

    manifest_fn = runtmp.output(f"mf.{ext}")
    manifest = BaseCollectionManifest.load_from_filename(manifest_fn)

    assert len(manifest) == 6
    md5_list = [row["md5"] for row in manifest.rows]
    assert "16869d2c8a1d29d1c8e56f5c561e585e" in md5_list
    assert "120d311cc785cc9d0df9dc0646b2b857" in md5_list
    assert "ea2a1ad233c2908529d124a330bcb672" in md5_list
    assert "bb0e6d90df01b7bd5d0956a5f9e3ed12" in md5_list
    assert "fbca5e5211e4d58427997fd5c8343e9a" in md5_list
    assert "1cbd888bf910f83ad8f1715509183223" in md5_list

    locations = set([row["internal_location"] for row in manifest.rows])
    assert protzip in locations
    assert hpzip in locations
    assert dayzip in locations
    assert len(locations) == 3, locations


def test_sig_collect_3_multiple_use_fromfile(runtmp, manifest_db_format):
    # collect a manifest from two .zip files using --from-file
    protzip = utils.get_test_data("prot/protein.zip")
    hpzip = utils.get_test_data("prot/hp.zip")
    dayzip = utils.get_test_data("prot/dayhoff.zip")

    ext = "sqlmf" if manifest_db_format == "sql" else "csv"

    fromfile = runtmp.output("fromfile.txt")
    with open(fromfile, "w") as fp:
        print(protzip, file=fp)
        print(hpzip, file=fp)
        print(dayzip, file=fp)

    runtmp.sourmash(
        "sig",
        "collect",
        "--from-file",
        "fromfile.txt",
        "-o",
        f"mf.{ext}",
        "-F",
        manifest_db_format,
    )

    manifest_fn = runtmp.output(f"mf.{ext}")
    manifest = BaseCollectionManifest.load_from_filename(manifest_fn)

    assert len(manifest) == 6
    md5_list = [row["md5"] for row in manifest.rows]
    assert "16869d2c8a1d29d1c8e56f5c561e585e" in md5_list
    assert "120d311cc785cc9d0df9dc0646b2b857" in md5_list
    assert "ea2a1ad233c2908529d124a330bcb672" in md5_list
    assert "bb0e6d90df01b7bd5d0956a5f9e3ed12" in md5_list
    assert "fbca5e5211e4d58427997fd5c8343e9a" in md5_list
    assert "1cbd888bf910f83ad8f1715509183223" in md5_list

    locations = set([row["internal_location"] for row in manifest.rows])
    assert protzip in locations
    assert hpzip in locations
    assert dayzip in locations
    assert len(locations) == 3, locations


def test_sig_collect_4_multiple_from_sig(runtmp, manifest_db_format):
    # collect a manifest from sig files
    sig43 = utils.get_test_data("47.fa.sig")
    sig63 = utils.get_test_data("63.fa.sig")

    ext = "sqlmf" if manifest_db_format == "sql" else "csv"

    runtmp.sourmash(
        "sig", "collect", sig43, sig63, "-o", f"mf.{ext}", "-F", manifest_db_format
    )

    manifest_fn = runtmp.output(f"mf.{ext}")
    manifest = BaseCollectionManifest.load_from_filename(manifest_fn)

    assert len(manifest) == 2
    md5_list = [row["md5"] for row in manifest.rows]
    assert "09a08691ce52952152f0e866a59f6261" in md5_list
    assert "38729c6374925585db28916b82a6f513" in md5_list

    locations = set([row["internal_location"] for row in manifest.rows])
    assert sig43 in locations
    assert sig63 in locations
    assert len(locations) == 2, locations


def test_sig_collect_4_multiple_from_sig_abspath(runtmp, manifest_db_format):
    # collect a manifest from sig files, forcing abspath
    sig43 = utils.get_test_data("47.fa.sig")
    sig63 = utils.get_test_data("63.fa.sig")

    shutil.copyfile(sig43, runtmp.output("47.fa.sig"))
    shutil.copyfile(sig63, runtmp.output("63.fa.sig"))

    ext = "sqlmf" if manifest_db_format == "sql" else "csv"

    runtmp.sourmash(
        "sig",
        "collect",
        "47.fa.sig",
        "63.fa.sig",
        "--abspath",
        "-o",
        f"mf.{ext}",
        "-F",
        manifest_db_format,
    )

    print(runtmp.last_result.out)
    print(runtmp.last_result.err)

    manifest_fn = runtmp.output(f"mf.{ext}")
    manifest = BaseCollectionManifest.load_from_filename(manifest_fn)

    assert len(manifest) == 2
    md5_list = [row["md5"] for row in manifest.rows]
    assert "09a08691ce52952152f0e866a59f6261" in md5_list
    assert "38729c6374925585db28916b82a6f513" in md5_list

    locations = set([row["internal_location"] for row in manifest.rows])
    print(locations)
    assert len(locations) == 2, locations

    for xx in locations:
        assert xx.startswith("/")


def test_sig_collect_4_multiple_no_abspath(runtmp, manifest_db_format):
    # collect a manifest from sig files, no abspath
    sig43 = utils.get_test_data("47.fa.sig")
    sig63 = utils.get_test_data("63.fa.sig")

    # copy files to tmp, where they will not have full paths
    shutil.copyfile(sig43, runtmp.output("47.fa.sig"))
    shutil.copyfile(sig63, runtmp.output("63.fa.sig"))

    ext = "sqlmf" if manifest_db_format == "sql" else "csv"

    runtmp.sourmash(
        "sig",
        "collect",
        "47.fa.sig",
        "63.fa.sig",
        "-o",
        f"mf.{ext}",
        "-F",
        manifest_db_format,
    )

    manifest_fn = runtmp.output(f"mf.{ext}")
    manifest = BaseCollectionManifest.load_from_filename(manifest_fn)

    assert len(manifest) == 2
    md5_list = [row["md5"] for row in manifest.rows]
    assert "09a08691ce52952152f0e866a59f6261" in md5_list
    assert "38729c6374925585db28916b82a6f513" in md5_list

    locations = set([row["internal_location"] for row in manifest.rows])
    print(locations)
    assert len(locations) == 2, locations
    assert "47.fa.sig" in locations
    assert "63.fa.sig" in locations

    runtmp.sourmash("sig", "cat", f"mf.{ext}")


def test_sig_collect_4_multiple_subdir_subdir_no_abspath(runtmp, manifest_db_format):
    # collect a manifest from sig files, no abspath; use a subdir for sketches
    # this should work with default behavior.
    sig43 = utils.get_test_data("47.fa.sig")
    sig63 = utils.get_test_data("63.fa.sig")

    # copy files to tmp, where they will not have full paths
    os.mkdir(runtmp.output("sigs_dir"))
    shutil.copyfile(sig43, runtmp.output("sigs_dir/47.fa.sig"))
    shutil.copyfile(sig63, runtmp.output("sigs_dir/63.fa.sig"))

    # put manifest in subdir too.
    os.mkdir(runtmp.output("mf_dir"))

    ext = "sqlmf" if manifest_db_format == "sql" else "csv"

    runtmp.sourmash(
        "sig",
        "collect",
        "sigs_dir/47.fa.sig",
        "sigs_dir/63.fa.sig",
        "-o",
        f"mf_dir/mf.{ext}",
        "-F",
        manifest_db_format,
        "--relpath",
    )

    manifest_fn = runtmp.output(f"mf_dir/mf.{ext}")
    manifest = BaseCollectionManifest.load_from_filename(manifest_fn)

    assert len(manifest) == 2
    md5_list = [row["md5"] for row in manifest.rows]
    assert "09a08691ce52952152f0e866a59f6261" in md5_list
    assert "38729c6374925585db28916b82a6f513" in md5_list

    locations = set([row["internal_location"] for row in manifest.rows])
    print(locations)
    assert len(locations) == 2, locations
    assert "../sigs_dir/47.fa.sig" in locations
    assert "../sigs_dir/63.fa.sig" in locations

    runtmp.sourmash("sig", "cat", f"mf_dir/mf.{ext}")


def test_sig_collect_4_multiple_cwd_subdir_no_abspath(runtmp, manifest_db_format):
    # collect a manifest from sig files, no abspath; use a subdir for sketches
    # this should work with default behavior.
    sig43 = utils.get_test_data("47.fa.sig")
    sig63 = utils.get_test_data("63.fa.sig")

    # copy files to tmp, where they will not have full paths
    os.mkdir(runtmp.output("sigs_dir"))
    shutil.copyfile(sig43, runtmp.output("sigs_dir/47.fa.sig"))
    shutil.copyfile(sig63, runtmp.output("sigs_dir/63.fa.sig"))

    ext = "sqlmf" if manifest_db_format == "sql" else "csv"

    runtmp.sourmash(
        "sig",
        "collect",
        "sigs_dir/47.fa.sig",
        "sigs_dir/63.fa.sig",
        "-o",
        f"mf.{ext}",
        "-F",
        manifest_db_format,
    )

    manifest_fn = runtmp.output(f"mf.{ext}")
    manifest = BaseCollectionManifest.load_from_filename(manifest_fn)

    assert len(manifest) == 2
    md5_list = [row["md5"] for row in manifest.rows]
    assert "09a08691ce52952152f0e866a59f6261" in md5_list
    assert "38729c6374925585db28916b82a6f513" in md5_list

    locations = set([row["internal_location"] for row in manifest.rows])
    print(locations)
    assert len(locations) == 2, locations
    assert "sigs_dir/47.fa.sig" in locations
    assert "sigs_dir/63.fa.sig" in locations

    runtmp.sourmash("sig", "cat", f"mf.{ext}")


def test_sig_collect_5_no_manifest_sbt_fail(runtmp, manifest_db_format):
    # collect a manifest from files that don't have one
    sbt_zip = utils.get_test_data("v6.sbt.zip")

    ext = "sqlmf" if manifest_db_format == "sql" else "csv"

    with pytest.raises(SourmashCommandFailed):
        runtmp.sourmash(
            "sig", "collect", sbt_zip, "-o", f"mf.{ext}", "-F", manifest_db_format
        )


def test_sig_collect_5_no_manifest_sbt_succeed(
    runtmp, manifest_db_format, abspath_relpath_v4
):
    # generate a manifest from files that don't have one when --no-require
    sbt_zip = utils.get_test_data("v6.sbt.zip")

    ext = "sqlmf" if manifest_db_format == "sql" else "csv"

    runtmp.sourmash(
        "sig",
        "collect",
        sbt_zip,
        "--no-require-manifest",
        "-o",
        f"mf.{ext}",
        "-F",
        manifest_db_format,
        abspath_relpath_v4,
    )

    manifest_fn = runtmp.output(f"mf.{ext}")
    manifest = BaseCollectionManifest.load_from_filename(manifest_fn)

    assert len(manifest) == 7
    locations = set([row["internal_location"] for row in manifest.rows])
    assert len(locations) == 1, locations
    assert sbt_zip in locations


def test_sig_collect_6_path_cwd_cwd(runtmp, manifest_db_format, abspath_relpath_v4):
    # check: manifest and sigs in cwd
    protzip = utils.get_test_data("prot/protein.zip")

    ext = "sqlmf" if manifest_db_format == "sql" else "csv"

    protzip_path = "protein.zip"
    shutil.copyfile(protzip, runtmp.output(protzip_path))

    mf_path = f"mf.{ext}"

    runtmp.sourmash(
        "sig",
        "collect",
        protzip_path,
        "-o",
        mf_path,
        "-F",
        manifest_db_format,
        abspath_relpath_v4,
    )

    runtmp.sourmash("sig", "cat", mf_path)


def test_sig_collect_6_path_cwd_subdir(runtmp, manifest_db_format, abspath_relpath_v4):
    # check: manifest in cwd, sigs in subdir
    protzip = utils.get_test_data("prot/protein.zip")

    ext = "sqlmf" if manifest_db_format == "sql" else "csv"

    os.mkdir(runtmp.output("sigs_dir"))
    protzip_path = "sigs_dir/protein.zip"
    shutil.copyfile(protzip, runtmp.output(protzip_path))

    mf_path = f"mf.{ext}"

    runtmp.sourmash(
        "sig",
        "collect",
        protzip_path,
        "-o",
        mf_path,
        "-F",
        manifest_db_format,
        abspath_relpath_v4,
    )

    runtmp.sourmash("sig", "cat", mf_path)


def test_sig_collect_6_path_subdir_cwd(runtmp, manifest_db_format, abspath_or_relpath):
    # check: manifest in cwd, sigs in subdir.  note, fails with default v4
    # behavior. see #3008.
    protzip = utils.get_test_data("prot/protein.zip")

    ext = "sqlmf" if manifest_db_format == "sql" else "csv"

    protzip_path = "protein.zip"
    shutil.copyfile(protzip, runtmp.output(protzip_path))

    os.mkdir(runtmp.output("mf_dir"))
    mf_path = f"mf_dir/mf.{ext}"

    runtmp.sourmash(
        "sig",
        "collect",
        protzip_path,
        "-o",
        mf_path,
        "-F",
        manifest_db_format,
        abspath_or_relpath,
    )

    runtmp.sourmash("sig", "cat", mf_path)


def test_sig_collect_6_path_subdir_subdir(
    runtmp, manifest_db_format, abspath_or_relpath
):
    # check: manifest and sigs in subdir.  note, fails with default v4
    # behavior. see #3008.
    protzip = utils.get_test_data("prot/protein.zip")

    ext = "sqlmf" if manifest_db_format == "sql" else "csv"

    os.mkdir(runtmp.output("sigs_dir"))
    protzip_path = "sigs_dir/protein.zip"
    shutil.copyfile(protzip, runtmp.output(protzip_path))

    os.mkdir(runtmp.output("mf_dir"))
    mf_path = f"mf_dir/mf.{ext}"

    runtmp.sourmash(
        "sig",
        "collect",
        protzip_path,
        "-o",
        mf_path,
        "-F",
        manifest_db_format,
        abspath_or_relpath,
    )

    runtmp.sourmash("sig", "cat", mf_path)



"""
Tests for the 'Index' class and protocol. All Index classes should support
this functionality.
"""

import pytest
import glob

import sourmash
from sourmash import SourmashSignature
from sourmash.index import (
    LinearIndex,
    ZipFileLinearIndex,
    LazyLinearIndex,
    MultiIndex,
    StandaloneManifestIndex,
    IndexSearchResult,
)
from sourmash.index import CounterGather
from sourmash.index.sqlite_index import SqliteIndex
from sourmash.index.revindex import RevIndex
from sourmash.sbt import SBT, GraphFactory
from sourmash.manifest import CollectionManifest, BaseCollectionManifest
from sourmash.lca.lca_db import LCA_Database, load_single_database
from sourmash.minhash import flatten_and_intersect_scaled, flatten_and_downsample_scaled

import sourmash_tst_utils as utils


def _load_three_sigs():
    # utility function - load & return these three sigs.
    sig2 = utils.get_test_data("2.fa.sig")
    sig47 = utils.get_test_data("47.fa.sig")
    sig63 = utils.get_test_data("63.fa.sig")

    ss2 = sourmash.load_one_signature(sig2, ksize=31)
    ss47 = sourmash.load_one_signature(sig47)
    ss63 = sourmash.load_one_signature(sig63)

    return [ss2, ss47, ss63]


def build_linear_index(runtmp):
    ss2, ss47, ss63 = _load_three_sigs()

    lidx = LinearIndex()
    lidx.insert(ss2)
    lidx.insert(ss47)
    lidx.insert(ss63)

    return lidx


def build_lazy_linear_index(runtmp):
    lidx = build_linear_index(runtmp)
    return LazyLinearIndex(lidx)


def build_sbt_index(runtmp):
    ss2, ss47, ss63 = _load_three_sigs()

    factory = GraphFactory(5, 100, 3)
    root = SBT(factory, d=2)

    root.insert(ss2)
    root.insert(ss47)
    root.insert(ss63)

    return root


def build_sbt_index_save_load(runtmp):
    root = build_sbt_index(runtmp)
    out = runtmp.output("xyz.sbt.zip")
    root.save(out)

    return sourmash.load_file_as_index(out)


def build_zipfile_index(runtmp):
    from sourmash.save_load import SaveSignatures_ZipFile

    location = runtmp.output("index.zip")
    with SaveSignatures_ZipFile(location) as save_sigs:
        for ss in _load_three_sigs():
            save_sigs.add(ss)

    idx = ZipFileLinearIndex.load(location)
    return idx


def build_multi_index(runtmp):
    siglist = _load_three_sigs()
    lidx = LinearIndex(siglist)

    mi = MultiIndex.load([lidx], [None], None)
    return mi


def build_standalone_manifest_index(runtmp):
    sig2 = utils.get_test_data("2.fa.sig")
    sig47 = utils.get_test_data("47.fa.sig")
    sig63 = utils.get_test_data("63.fa.sig")

    ss2 = sourmash.load_one_signature(sig2, ksize=31)
    ss47 = sourmash.load_one_signature(sig47)
    ss63 = sourmash.load_one_signature(sig63)

    siglist = [(ss2, sig2), (ss47, sig47), (ss63, sig63)]

    rows = []
    rows.extend((CollectionManifest.make_manifest_row(ss, loc) for ss, loc in siglist))
    mf = CollectionManifest(rows)
    mf_filename = runtmp.output("mf.csv")

    mf.write_to_filename(mf_filename)

    idx = StandaloneManifestIndex.load(mf_filename)
    return idx


def build_lca_index(runtmp):
    siglist = _load_three_sigs()
    db = LCA_Database(31, 1000, "DNA")
    for ss in siglist:
        db.insert(ss)

    return db


def build_lca_index_save_load(runtmp):
    db = build_lca_index(runtmp)
    outfile = runtmp.output("db.lca.json")
    db.save(outfile)

    return sourmash.load_file_as_index(outfile)


def build_sqlite_index(runtmp):
    filename = runtmp.output("idx.sqldb")
    db = SqliteIndex.create(filename)

    siglist = _load_three_sigs()
    for ss in siglist:
        db.insert(ss)

    return db


def build_revindex(runtmp):
    ss2, ss47, ss63 = _load_three_sigs()

    lidx = RevIndex(template=ss2.minhash)
    lidx.insert(ss2)
    lidx.insert(ss47)
    lidx.insert(ss63)

    return lidx


def build_lca_index_save_load_sql(runtmp):
    db = build_lca_index(runtmp)
    outfile = runtmp.output("db.lca.json")
    db.save(outfile, format="sql")

    x = load_single_database(outfile)
    db_load = x[0]

    return db_load


#
# create a fixture 'index_obj' that is parameterized by all of these
# building functions.
#


@pytest.fixture(
    params=[
        build_linear_index,
        build_lazy_linear_index,
        build_sbt_index,
        build_zipfile_index,
        build_multi_index,
        build_standalone_manifest_index,
        build_lca_index,
        build_sbt_index_save_load,
        build_lca_index_save_load,
        build_sqlite_index,
        build_lca_index_save_load_sql,
        #                        build_revindex,
    ]
)
def index_obj(request, runtmp):
    build_fn = request.param

    # build on demand
    return build_fn(runtmp)


###
### generic Index tests go here
###


def test_index_search_exact_match(index_obj):
    # search for an exact match
    ss2, ss47, ss63 = _load_three_sigs()

    sr = index_obj.search(ss2, threshold=1.0)
    print([s[1].name for s in sr])
    assert len(sr) == 1
    assert sr[0].signature.minhash == ss2.minhash
    assert sr[0].score == 1.0


def test_index_search_lower_threshold(index_obj):
    # search at a lower threshold/multiple results with ss47
    ss2, ss47, ss63 = _load_three_sigs()

    sr = index_obj.search(ss47, threshold=0.1)
    print([s[1].name for s in sr])
    assert len(sr) == 2
    sr.sort(key=lambda x: -x[0])
    assert sr[0].signature.minhash == ss47.minhash
    assert sr[0].score == 1.0
    assert sr[1].signature.minhash == ss63.minhash
    assert round(sr[1].score, 2) == 0.32


def test_index_search_lower_threshold_2(index_obj):
    # search at a lower threshold/multiple results with ss63
    ss2, ss47, ss63 = _load_three_sigs()

    sr = index_obj.search(ss63, threshold=0.1)
    print([s[1].name for s in sr])
    assert len(sr) == 2
    sr.sort(key=lambda x: -x[0])
    assert sr[0].signature.minhash == ss63.minhash
    assert sr[0].score == 1.0
    assert sr[1].signature.minhash == ss47.minhash
    assert round(sr[1].score, 2) == 0.32


def test_index_search_higher_threshold_2(index_obj):
    # search at a higher threshold/one match
    ss2, ss47, ss63 = _load_three_sigs()

    # search for sig63 with high threshold => 1 match
    sr = index_obj.search(ss63, threshold=0.8)
    print([s[1].name for s in sr])
    assert len(sr) == 1
    sr.sort(key=lambda x: -x[0])
    assert sr[0].signature.minhash == ss63.minhash
    assert sr[0].score == 1.0


def test_index_search_containment(index_obj):
    # search for containment at a low threshold/multiple results with ss63
    ss2, ss47, ss63 = _load_three_sigs()

    sr = index_obj.search(ss63, do_containment=True, threshold=0.1)
    print([s[1].name for s in sr])
    assert len(sr) == 2
    sr.sort(key=lambda x: -x[0])
    assert sr[0].signature.minhash == ss63.minhash
    assert sr[0].score == 1.0
    assert sr[1].signature.minhash == ss47.minhash
    assert round(sr[1].score, 2) == 0.48


def test_index_signatures(index_obj):
    # signatures works?
    siglist = list(index_obj.signatures())

    ss2, ss47, ss63 = _load_three_sigs()
    assert len(siglist) == 3

    # check md5sums, since 'in' doesn't always work
    md5s = set(ss.md5sum() for ss in siglist)
    assert ss2.md5sum() in md5s
    assert ss47.md5sum() in md5s
    assert ss63.md5sum() in md5s


def test_index_signatures_with_location(index_obj):
    # signatures_with_location works?
    siglist = list(index_obj.signatures_with_location())

    ss2, ss47, ss63 = _load_three_sigs()
    assert len(siglist) == 3

    # check md5sums, since 'in' doesn't always work
    md5s = set((ss.md5sum() for ss, loc in siglist))
    assert ss2.md5sum() in md5s
    assert ss47.md5sum() in md5s
    assert ss63.md5sum() in md5s


def test_index_len(index_obj):
    # len works?
    assert len(index_obj) == 3


def test_index_bool(index_obj):
    # bool works?
    assert bool(index_obj)


def test_index_location(index_obj):
    # location works?
    assert str(index_obj.location)


def test_index_manifest(index_obj):
    # manifest is either None or a BaseCollectionManifest
    manifest = index_obj.manifest
    if manifest is not None:
        assert isinstance(manifest, BaseCollectionManifest)


def test_index_select_basic(index_obj):
    # select does the basic thing ok
    idx = index_obj.select(
        ksize=31,
        moltype="DNA",
        abund=False,
        containment=True,
        scaled=1000,
        num=0,
        picklist=None,
    )

    assert len(idx) == 3
    siglist = list(idx.signatures())
    assert len(siglist) == 3

    # check md5sums, since 'in' doesn't always work
    md5s = set(ss.md5sum() for ss in siglist)
    ss2, ss47, ss63 = _load_three_sigs()
    assert ss2.md5sum() in md5s
    assert ss47.md5sum() in md5s
    assert ss63.md5sum() in md5s


def test_index_select_fail(index_obj):
    # non-int ksize - str
    with pytest.raises(ValueError):
        index_obj.select(ksize="31")

    # non-int ksize - float
    with pytest.raises(ValueError):
        index_obj.select(ksize=31.1)

    # case sensitive moltype
    with pytest.raises(ValueError):
        index_obj.select(moltype="dna")

    # unknown moltype
    with pytest.raises(ValueError):
        index_obj.select(moltype="foo")

    # non-int scaled
    with pytest.raises(ValueError):
        index_obj.select(scaled=1000.1)

    # non-int num
    with pytest.raises(ValueError):
        index_obj.select(num=1000.1)

    # non-bool abund
    with pytest.raises(ValueError):
        index_obj.select(abund=1)

    # extra parameters
    with pytest.raises(ValueError):
        index_obj.select(plausible_extra_parameter=5)


def test_index_select_nada(index_obj):
    # select works ok when nothing matches!

    # CTB: currently this EITHER raises a ValueError OR returns an empty
    # Index object, depending on implementation. :think:
    # See: https://github.com/sourmash-bio/sourmash/issues/1940
    try:
        idx = index_obj.select(ksize=21)
    except ValueError:
        idx = LinearIndex([])

    assert len(idx) == 0
    siglist = list(idx.signatures())
    assert len(siglist) == 0


def test_index_prefetch(index_obj):
    # test basic prefetch
    ss2, ss47, ss63 = _load_three_sigs()

    # search for ss2
    results = []
    for result in index_obj.prefetch(ss2, threshold_bp=0):
        results.append(result)

    assert len(results) == 1
    assert results[0].signature.minhash == ss2.minhash

    # search for ss47 - expect two results
    results = []
    for result in index_obj.prefetch(ss47, threshold_bp=0):
        results.append(result)

    assert len(results) == 2
    assert results[0].signature.minhash == ss47.minhash
    assert results[1].signature.minhash == ss63.minhash


def test_index_best_containment(index_obj):
    # test basic containment search
    ss2, ss47, ss63 = _load_three_sigs()

    match = index_obj.best_containment(ss2)
    assert match
    assert match.score == 1.0
    assert match.signature.minhash == ss2.minhash

    match = index_obj.best_containment(ss47)
    assert match
    assert match.score == 1.0
    assert match.signature.minhash == ss47.minhash


def test_index_best_containment_threshold_1(index_obj):
    # test best_containment() method, in some detail
    ss2, ss47, ss63 = _load_three_sigs()

    # now construct query signatures with specific numbers of hashes --
    # note, these signatures all have scaled=1000.

    mins = list(sorted(ss2.minhash.hashes))
    new_mh = ss2.minhash.copy_and_clear()

    # query with empty hashes
    assert not new_mh
    with pytest.raises(ValueError):
        index_obj.best_containment(SourmashSignature(new_mh))

    # add one hash
    new_mh.add_hash(mins.pop())
    assert len(new_mh) == 1

    result = index_obj.best_containment(SourmashSignature(new_mh))
    assert result
    containment, match_sig, name = result
    assert containment == 1.0
    assert match_sig.minhash == ss2.minhash

    # check with a threshold -> should be no results.
    with pytest.raises(ValueError):
        index_obj.best_containment(SourmashSignature(new_mh), threshold_bp=5000)

    # add three more hashes => length of 4
    new_mh.add_hash(mins.pop())
    new_mh.add_hash(mins.pop())
    new_mh.add_hash(mins.pop())
    assert len(new_mh) == 4

    result = index_obj.best_containment(SourmashSignature(new_mh))
    assert result
    containment, match_sig, name = result
    assert containment == 1.0
    assert match_sig.minhash == ss2.minhash

    # check with a too-high threshold -> should be no results.
    with pytest.raises(ValueError):
        index_obj.best_containment(SourmashSignature(new_mh), threshold_bp=5000)


def test_best_containment_threshold_5(index_obj):
    # test gather() method, in some detail
    ss2, ss47, ss63 = _load_three_sigs()

    # now construct query signatures with specific numbers of hashes --
    # note, these signatures all have scaled=1000.

    mins = list(sorted(ss2.minhash.hashes.keys()))
    new_mh = ss2.minhash.copy_and_clear()

    # add five hashes
    for i in range(5):
        new_mh.add_hash(mins.pop())
        new_mh.add_hash(mins.pop())
        new_mh.add_hash(mins.pop())
        new_mh.add_hash(mins.pop())
        new_mh.add_hash(mins.pop())

    # should get a result with no threshold (any match at all is returned)
    result = index_obj.best_containment(SourmashSignature(new_mh))
    assert result
    containment, match_sig, name = result
    assert containment == 1.0
    assert match_sig.minhash == ss2.minhash

    # now, check with a threshold_bp that should be meet-able.
    result = index_obj.best_containment(SourmashSignature(new_mh), threshold_bp=5000)
    assert result
    containment, match_sig, name = result
    assert containment == 1.0
    assert match_sig.minhash == ss2.minhash


###
### CounterGather tests
###


def create_basic_counter_gather(runtmp):
    "Construct a CounterGather class."
    return CounterGather


class CounterGather_LinearIndex:
    """
    Provides an (inefficient) CounterGather-style class, for
    protocol testing purposes.
    """

    def __init__(self, orig_query):
        "Constructor - take a SourmashSignature that is the original query."
        orig_query_mh = orig_query.minhash
        if orig_query_mh.scaled == 0:
            raise ValueError

        # Index object used to actually track matches.
        self.idx = LinearIndex()
        self.orig_query_mh = orig_query_mh.copy().flatten()
        self.query_started = 0
        self.scaled = orig_query_mh.scaled
        self.locations = {}

    def add(self, ss, *, location=None, require_overlap=True):
        "Insert potential match."
        if self.query_started:
            raise ValueError("cannot add more signatures to counter after peek/consume")

        # skip duplicates
        md5 = ss.md5sum()
        if md5 in self.locations:
            return

        # confirm that this match has an overlap...
        add_mh = ss.minhash.flatten()
        overlap = self.orig_query_mh.count_common(add_mh, downsample=True)

        # ...figure out what scaled we are operating at now...
        if overlap:
            self.downsample(add_mh.scaled)
        elif require_overlap:
            raise ValueError("no overlap between query and signature!?")

        # ...and add to the Index, while also tracking location!
        self.idx.insert(ss)
        self.locations[md5] = location

    def signatures(self):
        "Yield all signatures"
        return self.idx.signatures()

    def downsample(self, scaled):
        "Track highest scaled across all possible matches."
        if scaled > self.scaled:
            self.scaled = scaled
        return self.scaled

    def peek(self, cur_query_mh, *, threshold_bp=0):
        """
        Find best match to current query within this CounterGather object.
        """
        self.query_started = 1

        scaled = self.downsample(cur_query_mh.scaled)
        cur_query_mh = flatten_and_downsample_scaled(cur_query_mh, scaled)

        # no hashes remaining? exit.
        if not self.orig_query_mh or not cur_query_mh:
            return []

        # verify current query is a subset of the original.
        if cur_query_mh.contained_by(self.orig_query_mh, downsample=True) < 1:
            raise ValueError("current query not a subset of original query")

        # did we get a match?
        res = self.idx.peek(cur_query_mh, threshold_bp=threshold_bp)
        if not res:
            return []
        sr, intersect_mh = res

        # got match - replace location & return.
        match = sr.signature
        md5 = match.md5sum()
        location = self.locations[md5]
        return IndexSearchResult(sr.score, match, location), intersect_mh

    def consume(self, *args, **kwargs):
        self.query_started = 1
        return self.idx.consume(*args, **kwargs)


class CounterGather_LCA:
    """
    Provides an alternative implementation of a CounterGather-style class,
    based on LCA_Database. This is currently just for protocol
    and API testing purposes.
    """

    def __init__(self, query):
        from sourmash.lca.lca_db import LCA_Database

        query_mh = query.minhash
        if query_mh.scaled == 0:
            raise ValueError("must use scaled MinHash")

        self.orig_query_mh = query_mh
        lca_db = LCA_Database(query_mh.ksize, query_mh.scaled, query_mh.moltype)
        self.db = lca_db
        self.siglist = {}
        self.locations = {}
        self.query_started = 0

    def add(self, ss, *, location=None, require_overlap=True):
        "Add this signature into the counter."
        if self.query_started:
            raise ValueError("cannot add more signatures to counter after peek/consume")

        overlap = self.orig_query_mh.count_common(ss.minhash, True)
        if overlap:
            self.downsample(ss.minhash.scaled)
        elif require_overlap:
            raise ValueError("no overlap between query and signature!?")

        self.db.insert(ss)

        md5 = ss.md5sum()
        self.siglist[md5] = ss
        self.locations[md5] = location

    def signatures(self):
        "Yield all signatures."
        yield from self.siglist.values()

    def downsample(self, scaled):
        "Track highest scaled across all possible matches."
        if scaled > self.db.scaled:
            self.db.downsample_scaled(scaled)
        return self.db.scaled

    def peek(self, query_mh, *, threshold_bp=0):
        "Return next possible match."
        from sourmash import SourmashSignature

        self.query_started = 1
        scaled = self.downsample(query_mh.scaled)
        query_mh = query_mh.downsample(scaled=scaled)

        if not self.orig_query_mh or not query_mh:
            return []

        if query_mh.contained_by(self.orig_query_mh, downsample=True) < 1:
            raise ValueError("current query not a subset of original query")

        query_ss = SourmashSignature(query_mh)

        # returns search_result, intersect_mh
        try:
            result = self.db.best_containment(query_ss, threshold_bp=threshold_bp)
        except ValueError:
            result = None

        if not result:
            return []

        cont = result.score
        match = result.signature

        intersect_mh = flatten_and_intersect_scaled(result.signature.minhash, query_mh)

        md5 = result.signature.md5sum()
        location = self.locations[md5]

        new_sr = IndexSearchResult(cont, match, location)
        return [new_sr, intersect_mh]

    def consume(self, intersect_mh):
        self.query_started = 1


@pytest.fixture(
    params=[
        CounterGather,
        CounterGather_LinearIndex,
        CounterGather_LCA,
    ]
)
def counter_gather_constructor(request):
    build_fn = request.param

    # build on demand
    return build_fn


def test_counter_get_signatures(counter_gather_constructor):
    # test .signatures() method
    query_mh = sourmash.MinHash(n=0, ksize=31, scaled=1)
    query_mh.add_many(range(0, 20))
    query_ss = SourmashSignature(query_mh, name="query")

    match_mh_1 = query_mh.copy_and_clear()
    match_mh_1.add_many(range(0, 10))
    match_ss_1 = SourmashSignature(match_mh_1, name="match1")

    match_mh_2 = query_mh.copy_and_clear()
    match_mh_2.add_many(range(10, 15))
    match_ss_2 = SourmashSignature(match_mh_2, name="match2")

    match_mh_3 = query_mh.copy_and_clear()
    match_mh_3.add_many(range(15, 17))
    match_ss_3 = SourmashSignature(match_mh_3, name="match3")

    counter = counter_gather_constructor(query_ss)
    counter.add(match_ss_1)
    counter.add(match_ss_2)
    counter.add(match_ss_3)

    siglist = list(counter.signatures())
    assert len(siglist) == 3
    assert match_ss_1 in siglist
    assert match_ss_2 in siglist
    assert match_ss_3 in siglist


def _consume_all(query_mh, counter, threshold_bp=0):
    results = []
    query_mh = query_mh.to_mutable()

    last_intersect_size = None
    while 1:
        result = counter.peek(query_mh, threshold_bp=threshold_bp)
        if not result:
            break

        sr, intersect_mh = result
        print(sr.signature.name, len(intersect_mh))
        if last_intersect_size:
            assert len(intersect_mh) <= last_intersect_size

        last_intersect_size = len(intersect_mh)

        counter.consume(intersect_mh)
        query_mh.remove_many(intersect_mh.hashes)

        results.append((sr, len(intersect_mh)))

    return results


def test_counter_gather_1(counter_gather_constructor):
    # check a contrived set of non-overlapping gather results,
    # generated via CounterGather
    query_mh = sourmash.MinHash(n=0, ksize=31, scaled=1)
    query_mh.add_many(range(0, 20))
    query_ss = SourmashSignature(query_mh, name="query")

    match_mh_1 = query_mh.copy_and_clear()
    match_mh_1.add_many(range(0, 10))
    match_ss_1 = SourmashSignature(match_mh_1, name="match1")

    match_mh_2 = query_mh.copy_and_clear()
    match_mh_2.add_many(range(10, 15))
    match_ss_2 = SourmashSignature(match_mh_2, name="match2")

    match_mh_3 = query_mh.copy_and_clear()
    match_mh_3.add_many(range(15, 17))
    match_ss_3 = SourmashSignature(match_mh_3, name="match3")

    # load up the counter
    counter = counter_gather_constructor(query_ss)
    counter.add(match_ss_1)
    counter.add(match_ss_2)
    counter.add(match_ss_3)

    results = _consume_all(query_ss.minhash, counter)

    expected = (
        ["match1", 10],
        ["match2", 5],
        ["match3", 2],
    )
    assert len(results) == len(expected), results

    for (sr, size), (exp_name, exp_size) in zip(results, expected):
        sr_name = sr.signature.name.split()[0]

        assert sr_name == exp_name
        assert size == exp_size


def test_counter_gather_1_b(counter_gather_constructor):
    # check a contrived set of somewhat-overlapping gather results,
    # generated via CounterGather. Here the overlaps are structured
    # so that the gather results are the same as those in
    # test_counter_gather_1(), even though the overlaps themselves are
    # larger.
    query_mh = sourmash.MinHash(n=0, ksize=31, scaled=1)
    query_mh.add_many(range(0, 20))
    query_ss = SourmashSignature(query_mh, name="query")

    match_mh_1 = query_mh.copy_and_clear()
    match_mh_1.add_many(range(0, 10))
    match_ss_1 = SourmashSignature(match_mh_1, name="match1")

    match_mh_2 = query_mh.copy_and_clear()
    match_mh_2.add_many(range(7, 15))
    match_ss_2 = SourmashSignature(match_mh_2, name="match2")

    match_mh_3 = query_mh.copy_and_clear()
    match_mh_3.add_many(range(13, 17))
    match_ss_3 = SourmashSignature(match_mh_3, name="match3")

    # load up the counter
    counter = counter_gather_constructor(query_ss)
    counter.add(match_ss_1)
    counter.add(match_ss_2)
    counter.add(match_ss_3)

    results = _consume_all(query_ss.minhash, counter)

    expected = (
        ["match1", 10],
        ["match2", 5],
        ["match3", 2],
    )
    assert len(results) == len(expected), results

    for (sr, size), (exp_name, exp_size) in zip(results, expected):
        sr_name = sr.signature.name.split()[0]

        assert sr_name == exp_name
        assert size == exp_size


def test_counter_gather_1_c_with_threshold(counter_gather_constructor):
    # check a contrived set of somewhat-overlapping gather results,
    # generated via CounterGather. Here the overlaps are structured
    # so that the gather results are the same as those in
    # test_counter_gather_1(), even though the overlaps themselves are
    # larger.
    # use a threshold, here.

    query_mh = sourmash.MinHash(n=0, ksize=31, scaled=1)
    query_mh.add_many(range(0, 20))
    query_ss = SourmashSignature(query_mh, name="query")

    match_mh_1 = query_mh.copy_and_clear()
    match_mh_1.add_many(range(0, 10))
    match_ss_1 = SourmashSignature(match_mh_1, name="match1")

    match_mh_2 = query_mh.copy_and_clear()
    match_mh_2.add_many(range(7, 15))
    match_ss_2 = SourmashSignature(match_mh_2, name="match2")

    match_mh_3 = query_mh.copy_and_clear()
    match_mh_3.add_many(range(13, 17))
    match_ss_3 = SourmashSignature(match_mh_3, name="match3")

    # load up the counter
    counter = counter_gather_constructor(query_ss)
    counter.add(match_ss_1)
    counter.add(match_ss_2)
    counter.add(match_ss_3)

    results = _consume_all(query_ss.minhash, counter, threshold_bp=3)

    expected = (["match1", 10], ["match2", 5])
    assert len(results) == len(expected), results

    for (sr, size), (exp_name, exp_size) in zip(results, expected):
        sr_name = sr.signature.name.split()[0]

        assert sr_name == exp_name
        assert size == exp_size


def test_counter_gather_1_d_diff_scaled(counter_gather_constructor):
    # test as above, but with different scaled.
    query_mh = sourmash.MinHash(n=0, ksize=31, scaled=1)
    query_mh.add_many(range(0, 20))
    query_ss = SourmashSignature(query_mh, name="query")

    match_mh_1 = query_mh.copy_and_clear().downsample(scaled=10)
    match_mh_1.add_many(range(0, 10))
    match_ss_1 = SourmashSignature(match_mh_1, name="match1")

    match_mh_2 = query_mh.copy_and_clear().downsample(scaled=20)
    match_mh_2.add_many(range(7, 15))
    match_ss_2 = SourmashSignature(match_mh_2, name="match2")

    match_mh_3 = query_mh.copy_and_clear().downsample(scaled=30)
    match_mh_3.add_many(range(13, 17))
    match_ss_3 = SourmashSignature(match_mh_3, name="match3")

    # load up the counter
    counter = counter_gather_constructor(query_ss)
    counter.add(match_ss_1)
    counter.add(match_ss_2)
    counter.add(match_ss_3)

    results = _consume_all(query_ss.minhash, counter)

    expected = (
        ["match1", 10],
        ["match2", 5],
        ["match3", 2],
    )
    assert len(results) == len(expected), results

    for (sr, size), (exp_name, exp_size) in zip(results, expected):
        sr_name = sr.signature.name.split()[0]

        assert sr_name == exp_name
        assert size == exp_size


def test_counter_gather_1_d_diff_scaled_query(counter_gather_constructor):
    # test as above, but with different scaled for QUERY.
    query_mh = sourmash.MinHash(n=0, ksize=31, scaled=1)
    query_mh.add_many(range(0, 20))

    match_mh_1 = query_mh.copy_and_clear().downsample(scaled=10)
    match_mh_1.add_many(range(0, 10))
    match_ss_1 = SourmashSignature(match_mh_1, name="match1")

    match_mh_2 = query_mh.copy_and_clear().downsample(scaled=20)
    match_mh_2.add_many(range(7, 15))
    match_ss_2 = SourmashSignature(match_mh_2, name="match2")

    match_mh_3 = query_mh.copy_and_clear().downsample(scaled=30)
    match_mh_3.add_many(range(13, 17))
    match_ss_3 = SourmashSignature(match_mh_3, name="match3")

    # downsample query now -
    query_ss = SourmashSignature(query_mh.downsample(scaled=100), name="query")

    # load up the counter
    counter = counter_gather_constructor(query_ss)
    counter.add(match_ss_1)
    counter.add(match_ss_2)
    counter.add(match_ss_3)

    results = _consume_all(query_ss.minhash, counter)

    expected = (
        ["match1", 10],
        ["match2", 5],
        ["match3", 2],
    )
    assert len(results) == len(expected), results

    for (sr, size), (exp_name, exp_size) in zip(results, expected):
        sr_name = sr.signature.name.split()[0]

        assert sr_name == exp_name
        assert size == exp_size


def test_counter_gather_1_e_abund_query(counter_gather_constructor):
    # test as above, but abund query
    query_mh = sourmash.MinHash(n=0, ksize=31, scaled=1, track_abundance=1)
    query_mh.add_many(range(0, 20))
    query_ss = SourmashSignature(query_mh, name="query")

    match_mh_1 = query_mh.copy_and_clear().flatten()
    match_mh_1.add_many(range(0, 10))
    match_ss_1 = SourmashSignature(match_mh_1, name="match1")

    match_mh_2 = query_mh.copy_and_clear().flatten()
    match_mh_2.add_many(range(7, 15))
    match_ss_2 = SourmashSignature(match_mh_2, name="match2")

    match_mh_3 = query_mh.copy_and_clear().flatten()
    match_mh_3.add_many(range(13, 17))
    match_ss_3 = SourmashSignature(match_mh_3, name="match3")

    # load up the counter
    counter = counter_gather_constructor(query_ss)
    counter.add(match_ss_1)
    counter.add(match_ss_2)
    counter.add(match_ss_3)

    # must flatten before peek!
    results = _consume_all(query_ss.minhash.flatten(), counter)

    expected = (
        ["match1", 10],
        ["match2", 5],
        ["match3", 2],
    )
    assert len(results) == len(expected), results

    for (sr, size), (exp_name, exp_size) in zip(results, expected):
        sr_name = sr.signature.name.split()[0]

        assert sr_name == exp_name
        assert size == exp_size


def test_counter_gather_1_f_abund_match(counter_gather_constructor):
    # test as above, but abund query
    query_mh = sourmash.MinHash(n=0, ksize=31, scaled=1, track_abundance=1)
    query_mh.add_many(range(0, 20))
    query_ss = SourmashSignature(query_mh.flatten(), name="query")

    match_mh_1 = query_mh.copy_and_clear()
    match_mh_1.add_many(range(0, 10))
    match_ss_1 = SourmashSignature(match_mh_1, name="match1")

    match_mh_2 = query_mh.copy_and_clear()
    match_mh_2.add_many(range(7, 15))
    match_ss_2 = SourmashSignature(match_mh_2, name="match2")

    match_mh_3 = query_mh.copy_and_clear()
    match_mh_3.add_many(range(13, 17))
    match_ss_3 = SourmashSignature(match_mh_3, name="match3")

    # load up the counter
    counter = counter_gather_constructor(query_ss)
    counter.add(match_ss_1)
    counter.add(match_ss_2)
    counter.add(match_ss_3)

    # must flatten before peek!
    results = _consume_all(query_ss.minhash.flatten(), counter)

    expected = (
        ["match1", 10],
        ["match2", 5],
        ["match3", 2],
    )
    assert len(results) == len(expected), results

    for (sr, size), (exp_name, exp_size) in zip(results, expected):
        sr_name = sr.signature.name.split()[0]

        assert sr_name == exp_name
        assert size == exp_size


def test_counter_gather_2(counter_gather_constructor):
    # check basic set of gather results on semi-real data,
    # generated via CounterGather
    testdata_combined = utils.get_test_data("gather/combined.sig")
    testdata_glob = utils.get_test_data("gather/GCF*.sig")
    testdata_sigs = glob.glob(testdata_glob)

    query_ss = sourmash.load_one_signature(testdata_combined, ksize=21)
    subject_sigs = [
        (sourmash.load_one_signature(t, ksize=21), t) for t in testdata_sigs
    ]

    # load up the counter
    counter = counter_gather_constructor(query_ss)
    for ss, loc in subject_sigs:
        counter.add(ss, location=loc)

    results = _consume_all(query_ss.minhash, counter)

    expected = (
        ["NC_003198.1", 487],
        ["NC_000853.1", 192],
        ["NC_011978.1", 169],
        ["NC_002163.1", 157],
        ["NC_003197.2", 152],
        ["NC_009486.1", 92],
        ["NC_006905.1", 76],
        ["NC_011080.1", 59],
        ["NC_011274.1", 42],
        ["NC_006511.1", 31],
        ["NC_011294.1", 7],
        ["NC_004631.1", 2],
    )
    assert len(results) == len(expected)

    for (sr, size), (exp_name, exp_size) in zip(results, expected):
        sr_name = sr.signature.name.split()[0]
        print(sr_name, size)

        assert sr_name == exp_name
        assert size == exp_size


def test_counter_gather_exact_match(counter_gather_constructor):
    # query == match
    query_mh = sourmash.MinHash(n=0, ksize=31, scaled=1)
    query_mh.add_many(range(0, 20))
    query_ss = SourmashSignature(query_mh, name="query")

    # load up the counter; provide a location override, too.
    counter = counter_gather_constructor(query_ss)
    counter.add(query_ss, location="somewhere over the rainbow")

    results = _consume_all(query_ss.minhash, counter)
    assert len(results) == 1
    (sr, intersect_mh) = results[0]

    assert sr.score == 1.0
    assert sr.signature == query_ss
    assert sr.location == "somewhere over the rainbow"


def test_counter_gather_multiple_identical_matches(counter_gather_constructor):
    # test multiple identical matches being inserted, with only one return
    query_mh = sourmash.MinHash(n=0, ksize=31, scaled=1)
    query_mh.add_many(range(0, 20))
    query_ss = SourmashSignature(query_mh, name="query")

    # create counter...
    counter = counter_gather_constructor(query_ss)

    # now add multiple identical matches.
    match_mh = query_mh.copy_and_clear()
    match_mh.add_many(range(5, 15))

    for name in "match1", "match2", "match3":
        match_ss = SourmashSignature(match_mh, name=name)
        counter.add(match_ss, location=name)

    results = _consume_all(query_ss.minhash, counter)
    assert len(results) == 1

    sr, overlap_count = results[0]
    assert sr.score == 0.5
    assert overlap_count == 10

    # any one of the three is valid
    assert sr.location in ("match1", "match2", "match3")


def test_counter_gather_add_after_peek(counter_gather_constructor):
    # cannot add after peek or consume
    query_mh = sourmash.MinHash(n=0, ksize=31, scaled=1)
    query_mh.add_many(range(0, 20))
    query_ss = SourmashSignature(query_mh, name="query")

    # load up the counter
    counter = counter_gather_constructor(query_ss)
    counter.add(query_ss, location="somewhere over the rainbow")

    counter.peek(query_ss.minhash)

    with pytest.raises(ValueError):
        counter.add(query_ss, location="try again")


def test_counter_gather_add_after_consume(counter_gather_constructor):
    # cannot add after peek or consume
    query_mh = sourmash.MinHash(n=0, ksize=31, scaled=1)
    query_mh.add_many(range(0, 20))
    query_ss = SourmashSignature(query_mh, name="query")

    # load up the counter
    counter = counter_gather_constructor(query_ss)
    counter.add(query_ss, location="somewhere over the rainbow")

    counter.consume(query_ss.minhash)

    with pytest.raises(ValueError):
        counter.add(query_ss, location="try again")


def test_counter_gather_consume_empty_intersect(counter_gather_constructor):
    # check that consume works fine when there is an empty signature.
    query_mh = sourmash.MinHash(n=0, ksize=31, scaled=1)
    query_mh.add_many(range(0, 20))
    query_ss = SourmashSignature(query_mh, name="query")

    # load up the counter
    counter = counter_gather_constructor(query_ss)
    counter.add(query_ss, location="somewhere over the rainbow")

    # nothing really happens here :laugh:, just making sure there's no error
    counter.consume(query_ss.minhash.copy_and_clear())


def test_counter_gather_empty_initial_query(counter_gather_constructor):
    # check empty initial query
    query_mh = sourmash.MinHash(n=0, ksize=31, scaled=1)
    query_ss = SourmashSignature(query_mh, name="query")

    match_mh_1 = query_mh.copy_and_clear()
    match_mh_1.add_many(range(0, 10))
    match_ss_1 = SourmashSignature(match_mh_1, name="match1")

    # load up the counter
    counter = counter_gather_constructor(query_ss)
    counter.add(match_ss_1, require_overlap=False)

    assert counter.peek(query_ss.minhash) == []


def test_counter_gather_num_query(counter_gather_constructor):
    # check num query
    query_mh = sourmash.MinHash(n=500, ksize=31)
    query_mh.add_many(range(0, 10))
    query_ss = SourmashSignature(query_mh, name="query")

    with pytest.raises(ValueError):
        counter_gather_constructor(query_ss)


def test_counter_gather_empty_cur_query(counter_gather_constructor):
    # test empty cur query
    query_mh = sourmash.MinHash(n=0, ksize=31, scaled=1)
    query_mh.add_many(range(0, 20))
    query_ss = SourmashSignature(query_mh, name="query")

    # load up the counter
    counter = counter_gather_constructor(query_ss)
    counter.add(query_ss, location="somewhere over the rainbow")

    cur_query_mh = query_ss.minhash.copy_and_clear()
    results = _consume_all(cur_query_mh, counter)
    assert results == []


def test_counter_gather_add_num_matchy(counter_gather_constructor):
    # test add num query
    query_mh = sourmash.MinHash(n=0, ksize=31, scaled=1)
    query_mh.add_many(range(0, 20))
    query_ss = SourmashSignature(query_mh, name="query")

    match_mh = sourmash.MinHash(n=500, ksize=31)
    match_mh.add_many(range(0, 20))
    match_ss = SourmashSignature(match_mh, name="query")

    # load up the counter
    counter = counter_gather_constructor(query_ss)
    with pytest.raises(ValueError):
        counter.add(match_ss, location="somewhere over the rainbow")


def test_counter_gather_bad_cur_query(counter_gather_constructor):
    # test cur query that is not subset of original query
    query_mh = sourmash.MinHash(n=0, ksize=31, scaled=1)
    query_mh.add_many(range(0, 20))
    query_ss = SourmashSignature(query_mh, name="query")

    # load up the counter
    counter = counter_gather_constructor(query_ss)
    counter.add(query_ss, location="somewhere over the rainbow")

    cur_query_mh = query_ss.minhash.copy_and_clear()
    cur_query_mh.add_many(range(20, 30))
    with pytest.raises(ValueError):
        counter.peek(cur_query_mh)


def test_counter_gather_add_no_overlap(counter_gather_constructor):
    # check adding match with no overlap w/query
    query_mh = sourmash.MinHash(n=0, ksize=31, scaled=1)
    query_mh.add_many(range(0, 10))
    query_ss = SourmashSignature(query_mh, name="query")

    match_mh_1 = query_mh.copy_and_clear()
    match_mh_1.add_many(range(10, 20))
    match_ss_1 = SourmashSignature(match_mh_1, name="match1")

    # load up the counter
    counter = counter_gather_constructor(query_ss)
    with pytest.raises(ValueError):
        counter.add(match_ss_1)

    assert counter.peek(query_ss.minhash) == []


def test_counter_gather_big_threshold(counter_gather_constructor):
    # check 'peek' with a huge threshold
    query_mh = sourmash.MinHash(n=0, ksize=31, scaled=1)
    query_mh.add_many(range(0, 20))
    query_ss = SourmashSignature(query_mh, name="query")

    match_mh_1 = query_mh.copy_and_clear()
    match_mh_1.add_many(range(0, 10))
    match_ss_1 = SourmashSignature(match_mh_1, name="match1")

    # load up the counter
    counter = counter_gather_constructor(query_ss)
    counter.add(match_ss_1)

    # impossible threshold:
    threshold_bp = 30 * query_ss.minhash.scaled
    results = counter.peek(query_ss.minhash, threshold_bp=threshold_bp)
    assert results == []


def test_counter_gather_empty_counter(counter_gather_constructor):
    # check empty counter
    query_mh = sourmash.MinHash(n=0, ksize=31, scaled=1)
    query_ss = SourmashSignature(query_mh, name="query")

    # empty counter!
    counter = counter_gather_constructor(query_ss)

    assert counter.peek(query_ss.minhash) == []



"""display sourmash version and other information"""

import os
import screed
import sourmash
from sourmash.logging import notify
from sourmash.plugins import list_all_plugins


def subparser(subparsers):
    subparser = subparsers.add_parser("info")
    subparser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="report versions of khmer and screed",
    )


def info(verbose=False):
    "Report sourmash version + version of installed dependencies."
    notify(f"sourmash version {sourmash.VERSION}")
    notify(f"- loaded from path: {os.path.dirname(__file__)}")
    notify("")

    if verbose:
        notify("khmer version: None (internal Nodegraph)")
        notify("")

        notify(f"screed version {screed.__version__}")
        notify(f"- loaded from path: {os.path.dirname(screed.__file__)}")

        list_all_plugins()


def main(args):
    info(verbose=args.verbose)



"""check signature collections against a picklist"""

usage = """

    sourmash sig check <filenames> --picklist ... -o miss.csv -m manifest.csv

This will check the signature contents of <filenames> against the given
picklist, optionally outputting the unmatched picklist rows to 'miss.csv'
and optionally outputting a manifest of the matched signatures to
'manifest.csv'.

By default, 'sig check' requires a pre-existing manifest for collections;
this prevents potentially slow manifest rebuilding. You
can turn this check off with '--no-require-manifest'.

"""

from sourmash.cli.utils import (
    add_moltype_args,
    add_ksize_arg,
    add_picklist_args,
    add_pattern_args,
)


def subparser(subparsers):
    subparser = subparsers.add_parser("check", usage=usage)
    subparser.add_argument("signatures", nargs="*")
    subparser.add_argument(
        "-q", "--quiet", action="store_true", help="suppress non-error output"
    )
    subparser.add_argument(
        "-d", "--debug", action="store_true", help="provide debugging output"
    )
    subparser.add_argument(
        "-o",
        "--output-missing",
        metavar="FILE",
        help="output picklist with remaining unmatched entries to this file",
    )
    subparser.add_argument(
        "-f", "--force", action="store_true", help="try to load all files as signatures"
    )
    subparser.add_argument(
        "--from-file",
        help="a text file containing a list of files to load signatures from",
    )
    subparser.add_argument(
        "-m",
        "--save-manifest-matching",
        help="save a manifest of the matching entries to this file.",
    )
    subparser.add_argument(
        "--fail-if-missing",
        action="store_true",
        help="exit with an error code (-1) if there are any missing picklist values.",
    )
    subparser.add_argument(
        "--no-require-manifest",
        help="do not require a manifest; generate dynamically if needed",
        action="store_true",
    )
    subparser.add_argument(
        "-F",
        "--manifest-format",
        help="format of manifest output file; default is 'csv')",
        default="csv",
        choices=["csv", "sql"],
    )
    subparser.add_argument(
        "--abspath",
        "--use-absolute-paths",
        help="convert all locations to absolute paths",
        action="store_true",
    )
    subparser.add_argument(
        "--no-abspath",
        help="do not convert all locations to absolute paths",
        action="store_false",
        dest="abspath",
    )
    subparser.add_argument(
        "--relpath",
        "--use-relative-paths",
        help="convert all locations to paths relative to the output manifest",
        action="store_true",
    )
    subparser.add_argument(
        "--no-relpath",
        help="do not convert all locations to paths relative to the output manifest",
        action="store_false",
        dest="relpath",
    )

    add_ksize_arg(subparser)
    add_moltype_args(subparser)
    add_pattern_args(subparser)
    add_picklist_args(subparser)


def main(args):
    import sourmash

    return sourmash.sig.__main__.check(args)



"""
Utilities for jaccard/containment --> distance estimation
Equations from: https://github.com/KoslickiLab/mutation-rate-ci-calculator
Reference: https://doi.org/10.1101/2022.01.11.475870
"""

from dataclasses import dataclass, field
from scipy.optimize import brentq
from scipy.stats import norm as scipy_norm
from scipy.stats import binom
import numpy as np
from math import log, exp

from .logging import notify


def check_distance(dist):
    if not 0 <= dist <= 1:
        raise ValueError(f"Error: distance value {dist:.4f} is not between 0 and 1!")
    else:
        return dist


def check_prob_threshold(val, threshold=1e-3):
    """
    Check likelihood of no shared hashes based on chance alone (false neg).
    If too many exceed threshold, recommend user lower their scaled value.
    # !! when using this, keep count and recommend user lower scaled val
    """
    exceeds_threshold = False
    if threshold is not None and val > threshold:
        exceeds_threshold = True
    return val, exceeds_threshold


def check_jaccard_error(val, threshold=1e-4):
    exceeds_threshold = False
    if threshold is not None and val > threshold:
        exceeds_threshold = True
    return val, exceeds_threshold


@dataclass
class ANIResult:
    """Base class for distance/ANI from k-mer containment."""

    dist: float
    p_nothing_in_common: float
    p_threshold: float = 1e-3
    size_is_inaccurate: bool = False
    p_exceeds_threshold: bool = field(init=False)

    def check_dist_and_p_threshold(self):
        # check values
        self.dist = check_distance(self.dist)
        self.p_nothing_in_common, self.p_exceeds_threshold = check_prob_threshold(
            self.p_nothing_in_common, self.p_threshold
        )

    def __post_init__(self):
        self.check_dist_and_p_threshold()

    @property
    def ani(self):
        if self.size_is_inaccurate:
            return None
        return 1 - self.dist


@dataclass
class jaccardANIResult(ANIResult):
    """Class for distance/ANI from jaccard (includes jaccard_error)."""

    jaccard_error: float = None
    je_threshold: float = 1e-4

    def __post_init__(self):
        # check values
        self.check_dist_and_p_threshold()
        # check jaccard error
        if self.jaccard_error is not None:
            self.jaccard_error, self.je_exceeds_threshold = check_jaccard_error(
                self.jaccard_error, self.je_threshold
            )
        else:
            raise ValueError("Error: jaccard_error cannot be None.")

    @property
    def ani(self):
        # if jaccard error is too high (exceeds threshold), do not trust ANI estimate
        if self.je_exceeds_threshold or self.size_is_inaccurate:
            return None
        return 1 - self.dist


@dataclass
class ciANIResult(ANIResult):
    """
    Class for distance/ANI from containment: with confidence intervals.

    Set CI defaults to None, just in case CI can't be estimated for given sample.
    """

    dist_low: float = None
    dist_high: float = None

    def __post_init__(self):
        # check values
        self.check_dist_and_p_threshold()

        if self.dist_low is not None and self.dist_high is not None:
            self.dist_low = check_distance(self.dist_low)
            self.dist_high = check_distance(self.dist_high)

    @property
    def ani_low(self):
        if self.dist_high is None or self.size_is_inaccurate:
            return None
        return 1 - self.dist_high

    @property
    def ani_high(self):
        if self.dist_low is None or self.size_is_inaccurate:
            return None
        return 1 - self.dist_low


def r1_to_q(k, r1):
    r1 = float(r1)
    q = 1 - (1 - r1) ** k
    return float(q)


def var_n_mutated(L, k, r1, *, q=None):
    # there are computational issues in the variance formula that we solve here
    # by the use of higher-precision arithmetic; the problem occurs when r is
    # very small; for example, with L=10,k=2,r1=1e-6 standard precision
    # gives varN<0 which is nonsense; by using the mpf type, we get the correct
    # answer which is about 0.000038.
    if r1 == 0:
        return 0.0
    r1 = float(r1)
    if q is None:  # we assume that if q is provided, it is correct for r1
        q = r1_to_q(k, r1)
    varN = (
        L * (1 - q) * (q * (2 * k + (2 / r1) - 1) - 2 * k)
        + k * (k - 1) * (1 - q) ** 2
        + (2 * (1 - q) / (r1**2)) * ((1 + (k - 1) * (1 - q)) * r1 - q)
    )
    if varN < 0.0:  # this seems to happen only with super tiny test data
        raise ValueError("Error: varN <0.0!")
    return float(varN)


def exp_n_mutated(L, k, r1):
    q = r1_to_q(k, r1)
    return L * q


def exp_n_mutated_squared(L, k, p):
    return var_n_mutated(L, k, p) + exp_n_mutated(L, k, p) ** 2


def probit(p):
    return scipy_norm.ppf(p)


def handle_seqlen_nkmers(ksize, *, sequence_len_bp=None, n_unique_kmers=None):
    if n_unique_kmers is not None:
        return n_unique_kmers
    elif sequence_len_bp is None:
        # both are None, raise ValueError
        raise ValueError(
            "Error: distance estimation requires input of either 'sequence_len_bp' or 'n_unique_kmers'"
        )
    else:
        n_unique_kmers = sequence_len_bp - (ksize - 1)
        return n_unique_kmers


def set_size_chernoff(set_size, scaled, *, relative_error=0.05):
    """
    Computes the probability that the estimate: sketch_size * scaled deviates from the true
    set_size by more than relative_error. This relies on the fact that the sketch_size
    is binomially distributed with parameters sketch_size and 1/scale. The two-sided Chernoff
    bounds are used. This is depreciated in favor of set_size_exact_prob due to the later
    being accurate even for very small set sizes
    @param set_size: The number of distinct k-mers in the given set
    @param relative_error: the desired relative error (defaults to 5%)
    @return: float (the upper bound probability)
    """
    upper_bound = 1 - 2 * np.exp(-(relative_error**2) * set_size / (scaled * 3))
    return upper_bound


def set_size_exact_prob(set_size, scaled, *, relative_error=0.05):
    """
    Computes the exact probability that the estimate: sketch_size * scaled deviates from the true
    set_size by more than relative_error. This relies on the fact that the sketch_size
    is binomially distributed with parameters sketch_size and 1/scale. The CDF of the binomial distribution
    is used.
    @param set_size: The number of distinct k-mers in the given set
    @param relative_error: the desired relative error (defaults to 5%)
    @return: float (the upper bound probability)
    """
    # Need to check if the edge case is an integer or not. If not, don't include it in the equation
    pmf_arg = -set_size / scaled * (relative_error - 1)
    if pmf_arg == int(pmf_arg):
        prob = (
            binom.cdf(set_size / scaled * (relative_error + 1), set_size, 1 / scaled)
            - binom.cdf(-set_size / scaled * (relative_error - 1), set_size, 1 / scaled)
            + binom.pmf(-set_size / scaled * (relative_error - 1), set_size, 1 / scaled)
        )
    else:
        prob = binom.cdf(
            set_size / scaled * (relative_error + 1), set_size, 1 / scaled
        ) - binom.cdf(-set_size / scaled * (relative_error - 1), set_size, 1 / scaled)
    return prob


def get_expected_log_probability(n_unique_kmers, ksize, mutation_rate, scaled_fraction):
    """helper function
    Note that scaled here needs to be between 0 and 1
    (e.g. scaled 1000 --> scaled_fraction 0.001)
    """
    exp_nmut = exp_n_mutated(n_unique_kmers, ksize, mutation_rate)
    try:
        return (n_unique_kmers - exp_nmut) * log(1.0 - scaled_fraction)
    except:
        return float("-inf")


def get_exp_probability_nothing_common(
    mutation_rate, ksize, scaled, *, n_unique_kmers=None, sequence_len_bp=None
):
    """
    Given parameters, calculate the expected probability that nothing will be common
    between a fracminhash sketch of a original sequence and a fracminhash sketch of a mutated
    sequence. If this is above a threshold, we should suspect that the two sketches may have
    nothing in common. The threshold needs to be set with proper insights.

    Arguments: n_unique_kmers, ksize, mutation_rate, scaled
    Returns: float - expected likelihood that nothing is common between sketches
    """
    n_unique_kmers = handle_seqlen_nkmers(
        ksize, sequence_len_bp=sequence_len_bp, n_unique_kmers=n_unique_kmers
    )
    f_scaled = 1.0 / float(scaled)
    if mutation_rate == 1.0:
        return 1.0
    elif mutation_rate == 0.0:
        return 0.0
    return exp(
        get_expected_log_probability(n_unique_kmers, ksize, mutation_rate, f_scaled)
    )


def containment_to_distance(
    containment,
    ksize,
    scaled,
    *,
    n_unique_kmers=None,
    sequence_len_bp=None,
    confidence=0.95,
    estimate_ci=False,
    prob_threshold=1e-3,
):
    """
    Containment --> distance CI (one step)
    """
    sol1, sol2, point_estimate = None, None, None
    n_unique_kmers = handle_seqlen_nkmers(
        ksize, sequence_len_bp=sequence_len_bp, n_unique_kmers=n_unique_kmers
    )
    if containment == 0:
        # point_estimate = 1.0
        point_estimate = sol1 = sol2 = 1.0
    elif containment == 1:
        # point_estimate = 0.0
        point_estimate = sol1 = sol2 = 0.0
    else:
        point_estimate = 1.0 - containment ** (1.0 / ksize)
        if estimate_ci:
            try:
                alpha = 1 - confidence
                z_alpha = probit(1 - alpha / 2)
                f_scaled = (
                    1.0 / scaled
                )  # these use scaled as a fraction between 0 and 1

                bias_factor = 1 - (1 - f_scaled) ** n_unique_kmers

                term_1 = (1.0 - f_scaled) / (
                    f_scaled * n_unique_kmers**3 * bias_factor**2
                )

                def term_2(pest):
                    return n_unique_kmers * exp_n_mutated(
                        n_unique_kmers, ksize, pest
                    ) - exp_n_mutated_squared(n_unique_kmers, ksize, pest)

                def term_3(pest):
                    return (
                        var_n_mutated(n_unique_kmers, ksize, pest) / n_unique_kmers**2
                    )

                def var_direct(pest):
                    return term_1 * term_2(pest) + term_3(pest)

                def f1(pest):
                    return (
                        (1 - pest) ** ksize
                        + z_alpha * np.sqrt(var_direct(pest))
                        - containment
                    )

                def f2(pest):
                    return (
                        (1 - pest) ** ksize
                        - z_alpha * np.sqrt(var_direct(pest))
                        - containment
                    )

                sol1 = brentq(f1, 0.0000001, 0.9999999)
                sol2 = brentq(f2, 0.0000001, 0.9999999)

            except ValueError as exc:
                # afaict, this only happens with extremely small test data
                notify(
                    "WARNING: Cannot estimate ANI confidence intervals from containment. Do your sketches contain enough hashes?"
                )
                notify(str(exc))
                sol1 = sol2 = None

    # Do this here, so that we don't need to reconvert distance <--> identity later.
    prob_nothing_in_common = get_exp_probability_nothing_common(
        point_estimate, ksize, scaled, n_unique_kmers=n_unique_kmers
    )
    return ciANIResult(
        point_estimate,
        prob_nothing_in_common,
        dist_low=sol2,
        dist_high=sol1,
        p_threshold=prob_threshold,
    )


def jaccard_to_distance(
    jaccard,
    ksize,
    scaled,
    *,
    n_unique_kmers=None,
    sequence_len_bp=None,
    prob_threshold=1e-3,
    err_threshold=1e-4,
):
    """
    Given parameters, calculate point estimate for mutation rate from jaccard index.
    Uses formulas derived mathematically to compute the point estimate. The formula uses
    approximations, therefore a tiny error is associated with it. A lower bound of that error
    is also returned. A high error indicates that the point estimate cannot be trusted.
    Threshold of the error is open to interpretation, but suggested that > 10^-4 should be
    handled with caution.

    Note that the error is NOT a mutation rate, and therefore cannot be considered in
    something like mut.rate +/- error.

    Arguments: jaccard, ksize, scaled, n_unique_kmers
    # Returns: tuple (point_estimate_of_mutation_rate, lower_bound_of_error)

    # Returns: JaccardANIResult

    Note: point estimate does not consider impact of scaled, but p_nothing_in_common can be
    useful for determining whether scaled is sufficient for these comparisons.
    """
    error_lower_bound = None
    n_unique_kmers = handle_seqlen_nkmers(
        ksize, sequence_len_bp=sequence_len_bp, n_unique_kmers=n_unique_kmers
    )
    if jaccard == 0:
        point_estimate = 1.0
        error_lower_bound = 0.0
    elif jaccard == 1:
        point_estimate = 0.0
        error_lower_bound = 0.0
    else:
        point_estimate = 1.0 - (2.0 * jaccard / float(1 + jaccard)) ** (
            1.0 / float(ksize)
        )

        exp_n_mut = exp_n_mutated(n_unique_kmers, ksize, point_estimate)
        var_n_mut = var_n_mutated(n_unique_kmers, ksize, point_estimate)
        error_lower_bound = (
            1.0 * n_unique_kmers * var_n_mut / (n_unique_kmers + exp_n_mut) ** 3
        )
    prob_nothing_in_common = get_exp_probability_nothing_common(
        point_estimate, ksize, scaled, n_unique_kmers=n_unique_kmers
    )
    return jaccardANIResult(
        point_estimate,
        prob_nothing_in_common,
        jaccard_error=error_lower_bound,
        p_threshold=prob_threshold,
        je_threshold=err_threshold,
    )



"""Define the top-level command line interface for sourmash

This module handles user input when sourmash is invoked from the command line.
A top-level parser is defined for the `sourmash` command, and subparsers are
defined for each subcommand. Some sourmash operations are grouped together
using the `sourmash <subcmd> <subsubcmd>` pattern, and these are organized in
their own CLI submodules, each with a dedicated directory.
"""

from argparse import ArgumentParser, RawDescriptionHelpFormatter, SUPPRESS
import os
import sys

import sourmash

from . import utils

# Commands
from . import categorize
from . import compare
from . import compute
from . import gather
from . import import_csv
from . import info
from . import index
from . import migrate
from . import multigather
from . import plot
from . import prefetch
from . import sbt_combine
from . import search
from . import watch

# Subcommand groups
from . import lca
from . import sig
from . import sig as signature
from . import sketch
from . import storage
from . import tax
from . import scripts


class SourmashParser(ArgumentParser):
    _citation_printed = False

    def __init__(self, citation=True, **kwargs):
        super().__init__(**kwargs)
        self.citation = citation

    @classmethod
    def print_citation(cls):
        if cls._citation_printed:
            return
        from sourmash.logging import notify

        notify(f"\n== This is sourmash version {sourmash.VERSION}. ==")
        notify("== Please cite Irber et. al (2024), doi:10.21105/joss.06830. ==\n")
        cls._citation_printed = True

    def _subparser_from_name(self, name):
        """Given a name, get the subparser instance registered with this parser."""
        container = self._actions
        if name is None:
            return None
        for action in container:
            if action.choices is None:
                continue
            elif name in action.choices:
                return action.choices[name]

    def print_help(self):
        self.print_citation()
        super().print_help()

    def parse_args(self, args=None, namespace=None):
        if (args is None and len(sys.argv) == 1) or (
            args is not None and len(args) == 0
        ):
            self.print_help()
            raise SystemExit(1)
        args = super().parse_args(args=args, namespace=namespace)
        if ("quiet" not in args or not args.quiet) and self.citation:
            self.print_citation()

        if "subcmd" in args and args.subcmd is None:
            self._subparser_from_name(args.cmd).print_help()
            raise SystemExit(1)

        # BEGIN: dirty hacks to simultaneously support new and previous interface
        if hasattr(args, "subcmd") and args.subcmd == "import":
            args.subcmd = "ingest"
        # END: dirty hacks to simultaneously support new and previous interface
        return args


def get_parser():
    module_descs = {
        "tax": 'Integrate taxonomy information based on "gather" results',
        "lca": "Taxonomic operations",
        "sketch": "Create signatures",
        "sig": "Manipulate signature files",
        "storage": "Operations on storage",
        "scripts": "Plug-ins",
    }
    alias = {
        "sig": "signature",
        "ext": "scripts",
    }
    expert = set(
        ["categorize", "import_csv", "migrate", "multigather", "sbt_combine", "watch"]
    )

    clidir = os.path.dirname(__file__)
    basic_ops = utils.command_list(clidir)

    # provide a list of the basic operations - not expert, not submodules.
    user_ops = [op for op in basic_ops if op not in expert and op not in module_descs]
    usage = "    Basic operations\n"
    for op in user_ops:
        docstring = getattr(sys.modules[__name__], op).__doc__
        helpstring = f"sourmash {op:s} --help"
        usage += f"        {helpstring:25s} {docstring:s}\n"
    # next, all the subcommand ones - dive into subdirectories.
    cmd_group_dirs = next(os.walk(clidir))[1]
    cmd_group_dirs = filter(utils.opfilter, cmd_group_dirs)
    cmd_group_dirs = sorted(cmd_group_dirs)

    cmd_group_usage = [cmd for cmd in cmd_group_dirs if cmd not in alias.values()]
    for dirpath in cmd_group_usage:
        usage += "\n    " + module_descs[dirpath] + "\n"
        usage += f"        sourmash {dirpath:s} --help\n"
        if dirpath in alias:
            usage += f"        sourmash {alias[dirpath]:s} --help\n"

    desc = (
        "Create, compare, and manipulate k-mer sketches of biological sequences.\n\nUsage instructions:\n"
        + usage
    )
    parser = SourmashParser(
        prog="sourmash",
        description=desc,
        formatter_class=RawDescriptionHelpFormatter,
        usage=SUPPRESS,
    )
    parser._optionals.title = "Options"
    parser.add_argument(
        "-v", "--version", action="version", version="sourmash " + sourmash.VERSION
    )
    parser.add_argument(
        "-q", "--quiet", action="store_true", help="don't print citation information"
    )
    sub = parser.add_subparsers(
        title="Instructions",
        dest="cmd",
        metavar="cmd",
        help=SUPPRESS,
    )
    for op in basic_ops + cmd_group_dirs:
        getattr(sys.modules[__name__], op).subparser(sub)
    parser._action_groups.reverse()
    return parser


def parse_args(arglist=None):
    """
    Return an argparse 'args' object from parsing arglist.

    By default pulls arguments from sys.argv.

    Example usage:

    ```
    args = parse_args(['sig', 'filter', '-m', '10'])

    sourmash.sig.filter.__main__.filter(args)
    ```
    """
    return get_parser().parse_args(arglist)
